------------ Grid and ACE stuff ------------


local GridDynamicLayout = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0")
local Aura = AceLibrary("SpecialEvents-Aura-2.0")
local BS = AceLibrary("Babble-Spell-2.2")
local L = AceLibrary("AceLocale-2.2"):new("GridDynamicLayout")

L:RegisterTranslations("enUS", function()
	return {
		["Dynamic Layout"] = true,
		["Options for GridDynamicLayout."] = true,
		
		["ForcedHeader_title"] = "Force players into group",
		["ForcedTanks_desc"] = "Player names list separated by commas, those will always appear in the tank group",
		["ForcedHeals_desc"] = "Player names list separated by commas, those will always appear in the heal group",
		["ForcedRDPS_desc"] = "Player names list separated by commas, those will always appear in the ranged DPS group",
		["ForcedMDPS_desc"] = "Player names list separated by commas, those will always appear in the melee DPS group",
		["ForcedTanks_name"] = "Forced Tanks",
		["ForcedHeals_name"]="Forced Heals",
		["ForcedMDPS_name"]="Forced Melee DPS",
		["ForcedRDPS_name"]="Forced Ranged DPS",
		
		["TrashHeader_title"] = "Optional trash group",
		["ActivateAFKToTrash_name"] = "AFK",
		["ActivateAFKToTrash_desc"] = "If activated, any AFK user will be stored in the trash group at next layout update",
		["ActivateOfflineToTrash_name"] = "Offline",
		["ActivateOfflineToTrash_desc"] = "If activated, any Offline user will be stored in the trash group at next layout update",
		["ActivateDifferentZoneToTrash_name"] = "Players in a different zone",
		["ActivateDifferentZoneToTrash_desc"] = "If activated, any user who is not in the same zone you are will be stored in the trash group at next layout update",
		["ActivateNotVisibleToTrash_name"] = "Players too far away",
		["ActivateNotVisibleToTrash_desc"] = "If activated, any user who are too far away from you to be rendered will be stored in the trash group at next layout update",
		
		["MiscHeader_title"] = "Miscellaneous",
		["DefaultDPSWars_name"] = "Defaulted DPS warriors",
		["DefaultDPSWars_desc"] = "Player names list separated by commas, those will be defaulted as DPS warriors",
		["ActivateStats_name"] = "Activate statistical analysis",
		["ActivateStats_desc"] = "Mod will attempt to assign a default behaviour to user based on active buff / form to improve performances",
	}
end)

GridDynamicLayout = Grid:NewModule("GridDynamicLayout")
GridDynamicLayout.menuName = L["Dynamic Layout"]


------------ Locals ------------

local GROUPINDEX_TANK = 1
local GROUPINDEX_HEAL = 2
local GROUPINDEX_RDPS = 3
local GROUPINDEX_MDPS = 4
local GROUPINDEX_TRASH = 5

local UNITS_PER_COLUMN = 5
local MAX_COLUMNS = 8

local dynamicLayout = {
	defaults = {
		unitsPerColumn = UNITS_PER_COLUMN,
		maxColumns = MAX_COLUMNS,
		columnAnchorPoint = "LEFT",
		sortMethod="INDEX",
	},
	[GROUPINDEX_TANK] = {					 -- tanks
		nameList = "",
	},
	[GROUPINDEX_HEAL] = {					 -- healers
		nameList = "",
	},
	[GROUPINDEX_RDPS] = {					 -- ranged DPS
		nameList = "",
	},
	[GROUPINDEX_MDPS] = {					 -- melee DPS
		nameList = "",
	},
}

local dpsWarList = {
}

local MAX_SPECSCORE=6
local MIN_SIGNIFICANT_SPECSCORE=2
local UPDATE_DELAY = 1

------------ GridDynamicLayoutACE Options -------------

GridDynamicLayout.options = {
	ForcedHeader = {
		type = "header",
		name = L["ForcedHeader_title"],
		order = 10,
	},
	ForcedTanks = {
		type = "text",
		name = L["ForcedTanks_name"],
		desc = L["ForcedTanks_desc"],
		get = function() return GridDynamicLayout:generateOptionValue(GridDynamicLayout.db.profile.tankList) end,
		usage = L["ForcedTanks_desc"],
		set = function(v)
			GridDynamicLayout.db.profile.tankList = GridDynamicLayout:parseOptionValue(v)
			GridDynamicLayout:setUpdateNeeded("Forced Tank update")
		end,
		order = 11,
	},
	ForcedHeals = {
		type = "text",
		name = L["ForcedHeals_name"],
		desc = L["ForcedHeals_desc"],
		get = function() return GridDynamicLayout:generateOptionValue(GridDynamicLayout.db.profile.healList) end,
		usage = L["ForcedHeals_desc"],
		set = function(v)
			GridDynamicLayout.db.profile.healList = GridDynamicLayout:parseOptionValue(v)
			GridDynamicLayout:setUpdateNeeded("Forced Heal update")
		end,
		order = 12,
	},
	ForcedMDPS = {
		type = "text",
		name = L["ForcedMDPS_name"],
		desc = L["ForcedMDPS_desc"],
		get = function() return GridDynamicLayout:generateOptionValue(GridDynamicLayout.db.profile.MDPSList) end,
		usage = L["ForcedMDPS_desc"],
		set = function(v)
			GridDynamicLayout.db.profile.MDPSList = GridDynamicLayout:parseOptionValue(v)
			GridDynamicLayout:setUpdateNeeded("Forced melee DPS update")
		end,
		order = 13,
	},
	ForcedRDPS = {
		type = "text",
		name = L["ForcedRDPS_name"],
		desc = L["ForcedRDPS_desc"],
		get = function() return GridDynamicLayout:generateOptionValue(GridDynamicLayout.db.profile.RDPSList) end,
		usage = L["ForcedRDPS_name"],
		set = function(v)
			GridDynamicLayout.db.profile.RDPSList = GridDynamicLayout:parseOptionValue(v)
			GridDynamicLayout:setUpdateNeeded("Forced range DPS update")
		end,
		order = 14,
	},
	
	TrashHeader = {
		type = "header",
		name = L["TrashHeader_title"],
		order = 20,
	},
	ActivateAFKToTrash = {
		type = "toggle",
		name = L["ActivateAFKToTrash_name"],
		desc = L["ActivateAFKToTrash_desc"],
		get = function() return GridDynamicLayout.db.profile.activateAFKToTrash end,
		set = function(v)
			GridDynamicLayout.db.profile.activateAFKToTrash = v
			GridDynamicLayout:setUpdateNeeded("Toggle AFK to trash")	
		end,
		order = 21,
	},
	ActivateOfflineToTrash = {
		type = "toggle",
		name = L["ActivateOfflineToTrash_name"],
		desc = L["ActivateOfflineToTrash_desc"],
		get = function() return GridDynamicLayout.db.profile.activateOfflineToTrash end,
		set = function(v)
			GridDynamicLayout.db.profile.activateOfflineToTrash = v
			GridDynamicLayout:setUpdateNeeded("Toggle Offline to trash")	
		end,
		order = 22,
	},
	ActivateDifferentZoneToTrash = {
		type = "toggle",
		name = L["ActivateDifferentZoneToTrash_name"],
		desc = L["ActivateDifferentZoneToTrash_desc"],
		get = function() return GridDynamicLayout.db.profile.activateDifferentZoneToTrash end,
		set = function(v)
			GridDynamicLayout.db.profile.activateDifferentZoneToTrash = v
			GridDynamicLayout:setUpdateNeeded("Toggle Different Zone to trash")	
		end,
		order = 23,
	},
	ActivateNotVisibleToTrash = {
		type = "toggle",
		name = L["ActivateNotVisibleToTrash_name"],
		desc = L["ActivateNotVisibleToTrash_desc"],
		get = function() return GridDynamicLayout.db.profile.activateNotVisibleToTrash end,
		set = function(v)
			GridDynamicLayout.db.profile.activateNotVisibleToTrash = v
			GridDynamicLayout:setUpdateNeeded("Toggle Not visible to trash")	
		end,
		order = 24,
	},
	
	MiscHeader = {
		type = "header",
		name = L["MiscHeader_title"],
		order = 30,
	},
	DefaultDPSWar = {
		type = "text",
		name = L["DefaultDPSWars_name"],
		desc = L["DefaultDPSWars_desc"],
		get = function() return GridDynamicLayout:generateOptionValue(GridDynamicLayout.db.profile.defaultDPSWarList) end,
		usage = L["DefaultDPSWars_desc"],
		set = function(v)
			GridDynamicLayout.db.profile.defaultDPSWarList = GridDynamicLayout:parseOptionValue(v)
		end,
		order = 31,
	},
	ActivateStats = {
		type = "toggle",
		name = L["ActivateStats_name"],
		desc = L["ActivateStats_desc"],
		get = function() return GridDynamicLayout.db.profile.activateStats end,
		usage = L["ActivateStats_desc"],
		set = function(v)
			GridDynamicLayout.db.profile.activateStats = v
		end,
		order = 32,
	},
	--[[DumpStats = {
		type = "execute",
		name = "Dump Stats DB",
		desc = "Dump Stats DB",
		handler = GridDynamicLayout,
		func = "dumpStatsDB",
		order = 13,
	},]]--
}

function GridDynamicLayout:parseOptionValue(s)
	local result = {}
	if string.find(s, ",") then
		for a in gmatch(s..",", "(.-),") do
			if a ~= "" then
				result[a] = true
			end
		end
	elseif s ~= "" then
		result[s] = true
	end
	return result
end

function GridDynamicLayout:generateOptionValue(list)
	local result = ""
	for key,value in pairs(list) do
		if value then
			result = result .. key .. ","
		end
	end
	return result
end

-- Hook into GridLayout options
GridLayout.options.args["DynamicLayout"] = {
	type = "group",
	name = L["Dynamic Layout"],
	desc = L["Options for GridDynamicLayout."],
	disabled = InCombatLockdown,
	args = GridDynamicLayout.options
}

------------ GridDynamicLayoutDefaultDB -------------

GridDynamicLayout.defaultDB = {
	debug=false,
	activateStats=false,
	activateAFKToTrash=false,
	activateOfflineToTrash=false,
	activateDifferentZoneToTrash=false,
	activateNotVisibleToTrash=false,
	tankList = {},
	healList = {},
	MDPSList = {},
	RDPSList = {},
	defaultDPSWarList = {},
}

------------ GridDynamicLayout ------------

function GridDynamicLayout:OnInitialize()
	self.super.OnInitialize(self)
	-- Initializes session-specific data
	dpsWarList = self.db.profile.defaultDPSWarList
	self.ExitingCombatScheduled = false
	self.pertinentEventsRegistered = false
	self.updateNeeded = false
	self.stats = {}
end

function GridDynamicLayout:OnEnable()
	self:RegisterEvent("Grid_JoinedBattleground", "initializeLayoutJoinedBG")
	self:RegisterEvent("Grid_JoinedRaid", "initializeLayoutJoinedRaid")
	self:RegisterEvent("Grid_JoinedParty", "initializeLayoutJoinedParty")
	self:RegisterEvent("Grid_LeftParty", "unregisterPertinentEvents")
	self:ScheduleRepeatingEvent("GridDynamicLayout_CheckUpdate", UPDATE_DELAY)
	self:RegisterEvent("GridDynamicLayout_CheckUpdate", "initializeLayout")
end

function GridDynamicLayout:initializeLayoutJoinedBG()
	self:initializeLayoutJoined("Joined BG")
end

function GridDynamicLayout:initializeLayoutJoinedParty()
	self:initializeLayoutJoined("Joined Party")
end

function GridDynamicLayout:initializeLayoutJoinedRaid()
	self:initializeLayoutJoined("Joined Raid")
end

function GridDynamicLayout:initializeLayoutJoined(reason)
	self:setUpdateNeeded(reason)
	self:registerPertinentEvents()
end


function GridDynamicLayout:unregisterPertinentEvents()
	if self.pertinentEventsRegistered then
		self:UnregisterEvent("Grid_UnitLeft")
		self:UnregisterEvent("Grid_UnitJoined")
		self:UnregisterEvent("SpecialEvents_UnitBuffLost")
		self:UnregisterEvent("SpecialEvents_UnitBuffGained")
		self:UnregisterEvent("CHAT_MSG_SPELL_PERIODIC_PARTY_BUFFS")
		self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	end
	self.pertinentEventsRegistered = false
end

function GridDynamicLayout:registerPertinentEvents()
	if not self.pertinentEventsRegistered then
		self:RegisterEvent("Grid_UnitLeft", "initializeLayoutLeft")
		self:RegisterEvent("Grid_UnitJoined", "initializeLayoutJoin")
		self:RegisterEvent("SpecialEvents_UnitBuffLost", "checkBuff")
		self:RegisterEvent("SpecialEvents_UnitBuffGained", "checkBuff")
		self:RegisterEvent("CHAT_MSG_SPELL_PERIODIC_PARTY_BUFFS", "handleWarStanceChange")
		self:RegisterEvent("PLAYER_REGEN_ENABLED", "UpdateStatisticalAnalysis")
	end
	self.pertinentEventsRegistered = true
end

function GridDynamicLayout:initializeLayoutLeft(name)
	self:setUpdateNeeded("Unit Left - "..name)
end

function GridDynamicLayout:initializeLayoutJoin(name, unitid)
	self:setUpdateNeeded("Unit Join - "..name.." ("..unitid..")")
end

function GridDynamicLayout:handleWarStanceChange(msg)
	local unitName, buff
	for unitName, buff in string.gmatch(msg, "(.+) gains (.+).") do
		
		if unitName and buff then
			if unitName=="You" then unitName = UnitName(player) end
			if buff == BS["Berserker Stance"] or buff == BS["Battle Stance"] then
				dpsWarList[unitName]=true
				self:setUpdateNeeded("War stance change - "..unitName.." - to DPS")
			elseif buff == BS["Defensive Stance"] then
				dpsWarList[unitName]=false
				self:setUpdateNeeded("War stance change - "..unitName.." - to tank")
			end 
		end
	end
end

function GridDynamicLayout:getRealmedName(unit)
	local name, realm = UnitName(unit)
	
	if realm ~= nil and realm ~= "" then
		name = name .. "-" .. realm
	end
	
	return name
end

function GridDynamicLayout:addToGroup(grpNumber, unit)
	local name = self:getRealmedName(unit)
	if grpNumber==nil then
		self:Debug(name .. " (".. unit .. ") : could not find matching group" )
	elseif not dynamicLayout[grpNumber] then
		self:Debug("Attempted to add " .. name .. " (".. unit .. ") to non existing group number : " .. grpNumber )
	else
		dynamicLayout[grpNumber].nameList = dynamicLayout[grpNumber].nameList .. name .. ","
	end
end

function GridDynamicLayout:setUpdateNeeded(reason)
	self:Debug("Update needed : " .. reason)
	self.updateNeeded = true
end

function GridDynamicLayout:initializeLayout()
	if not self.updateNeeded then 
		return 
	end
	
	self:Debug("----- Initializing Layout ------")
	
	local i
	
	-- create or delete trash group as needed
	if self.db.profile.activateAFKToTrash or 
		self.db.profile.activateOfflineToTrash or 
		self.db.profile.activateDifferentZoneToTrash or 
		self.db.profile.activateNotVisibleToTrash
	then
		if dynamicLayout[GROUPINDEX_TRASH] == nil then
			dynamicLayout[GROUPINDEX_TRASH] = {
				nameList = "",
			}
		end
	else
		if dynamicLayout[GROUPINDEX_TRASH] ~= nil then
			dynamicLayout[GROUPINDEX_TRASH] = nil
		end
	end
	
	for _,v in pairs(dynamicLayout) do
		if v.nameList then
			v.nameList = ""
		end
	end
	
	-- Sorts raid by class / name
	sortedRaid = {}
	for i=1,40 do
		if UnitExists("raid"..i) then
			table.insert(sortedRaid, i)
		end
	end
	table.sort(sortedRaid, function(a, b)
		local _, aClass = UnitClass("raid"..a)
		local _, bClass = UnitClass("raid"..b)
		if not aClass~=nil and bClass~=nil then
			if aClass < bClass then
				return true
			elseif aClass == bClass and UnitName("raid"..a) < UnitName("raid"..b) then
				return true
			else
				return false
			end
		end
		return false
	end)
	
	-- Scans the raid and place each unit in appropriate nameList
	self:Debug("--------------")
	for i=1,#sortedRaid do
		self:Debug(sortedRaid[i]..">"..UnitName("raid"..sortedRaid[i]).."-"..UnitClass("raid"..sortedRaid[i]))
		if UnitExists("raid"..sortedRaid[i]) then
			self:addToGroup(self:getLayoutGroupIndex(sortedRaid[i]), "raid"..sortedRaid[i])
		end
	end
	self:Debug("--------------")
		
	
	if GridLayout.db.profile.layout == "Dynamic Layout" then
		if not InCombatLockdown() then
			GridLayout:ReloadLayout()
		elseif not self.ExitingCombatScheduled then
			self:ScheduleLeaveCombatAction(GridDynamicLayout, "ExitingCombat")
			self.ExitingCombatScheduled = true
		end
	end
	self.updateNeeded = false
end

function GridDynamicLayout:ExitingCombat()
	self:ScheduleLeaveCombatAction(GridLayout, "ReloadLayout")
	self.ExitingCombatScheduled = false
end

function GridDynamicLayout:getLayoutGroupIndex(i)
	local unit = "raid"..i
	-- Add to trash group if not in right subzone / is offline
	local _, _, _, _, _, _, zone, online = GetRaidRosterInfo(i)
	if 
		(self.db.profile.activateAFKToTrash and UnitIsAFK(unit)) or 
		(self.db.profile.activateOfflineToTrash and not online) or 
		(self.db.profile.activateDifferentZoneToTrash and zone~=GetRealZoneText()) or 
		(self.db.profile.activateNotVisibleToTrash and not UnitIsVisible(unit))
	then
		return GROUPINDEX_TRASH
	-- Add to tanks if "by name" setting used
	elseif self.db.profile.tankList[UnitName(unit)] then
		return GROUPINDEX_TANK
	-- Add to healers if "by name" setting used
	elseif self.db.profile.healList[UnitName(unit)] then
		return GROUPINDEX_HEAL
	-- Add to ranged DPS if "by name" setting used
	elseif self.db.profile.RDPSList[UnitName(unit)] then
		return GROUPINDEX_RDPS
	-- Add to melee DPS if "by name" setting used
	elseif self.db.profile.MDPSList[UnitName(unit)] then
		return GROUPINDEX_MDPS
	-- Use class / buff setting to assign role
	else
		local class
		_, class = UnitClass(unit)
		
		-- Add Shamans and Paladins to heal group
		if class=="SHAMAN" then
			return GROUPINDEX_HEAL
		-- Add Hunters, Mages and Warlocks to ranged DPS group
		elseif class=="HUNTER" or class=="MAGE" or class=="WARLOCK" then
			return GROUPINDEX_RDPS
		-- Add Rogues to melee DPS group
		elseif class=="ROGUE" then
			return GROUPINDEX_MDPS
		elseif class=="PALADIN" then
			return self:getLayoutGroupIndex_Paladin(unit)
		elseif class=="WARRIOR" then
			-- Handling of warriors (cannot detect stance by normal means) --
			if dpsWarList[UnitName(unit)] then
				return GROUPINDEX_MDPS
			else
				return GROUPINDEX_TANK
			end
		elseif class=="PRIEST" then
			return self:getLayoutGroupIndex_Priest(unit)
		elseif class=="DRUID" then
			return self:getLayoutGroupIndex_Druid(unit)
		end
	end
end

function GridDynamicLayout:getLayoutGroupIndex_Paladin(unit)
	local assumedSpec = nil
	local useStatData = false
	if self.db.profile.activateStats and self.stats[self:getRealmedName(unit)] then
		assumedSpec, useStatData = self:getAssumedSpec(unit)
	end
	
	if useStatData then
		if assumedSpec == "prot" then
			return GROUPINDEX_TANK
		else
			return GROUPINDEX_HEAL
		end
	else
		if Aura:UnitHasBuff(unit, BS["Righteous Fury"]) or Aura:UnitHasBuff(unit, BS["Improved Righteous Fury"]) then
			return GROUPINDEX_TANK
		else
			return GROUPINDEX_HEAL
		end
	end
end

function GridDynamicLayout:getLayoutGroupIndex_Priest(unit)
	local assumedSpec = nil
	local useStatData = false
	if self.db.profile.activateStats and self.stats[self:getRealmedName(unit)] then
		assumedSpec, useStatData = self:getAssumedSpec(unit)
	end

	if useStatData then
		if assumedSpec == "shadow" then
			return GROUPINDEX_RDPS
		else
			return GROUPINDEX_HEAL
		end
	else
		if Aura:UnitHasBuff(unit, BS["Shadowform"]) then
			return GROUPINDEX_RDPS
		else
			return GROUPINDEX_HEAL
		end
	end
end

function GridDynamicLayout:getLayoutGroupIndex_Druid(unit)
	local assumedSpec = nil
	local useStatData = false
	if self.db.profile.activateStats and self.stats[self:getRealmedName(unit)] then
		assumedSpec, useStatData = self:getAssumedSpec(unit)
	end
	
	if useStatData then
		if assumedSpec == "feral" then
			-- If detected as feral choices are cat / bear
			-- if cat put in melee dps, otherwise (even if in caster form) in tank group
			if Aura:UnitHasBuff(unit, BS["Cat Form"]) then
				return GROUPINDEX_MDPS
			else
				return GROUPINDEX_TANK
			end
		elseif assumedSpec == "balance" then
			return GROUPINDEX_RDPS
		else
			return GROUPINDEX_HEAL
		end
	else
		if Aura:UnitHasBuff(unit, BS["Dire Bear Form"]) or Aura:UnitHasBuff(unit, BS["Bear Form"]) then
			return GROUPINDEX_TANK
		elseif Aura:UnitHasBuff(unit, BS["Cat Form"]) then
			return GROUPINDEX_MDPS
		elseif Aura:UnitHasBuff(unit, BS["Moonkin Form"]) then
			return GROUPINDEX_RDPS
		else
			return GROUPINDEX_HEAL
		end
	end
end

function GridDynamicLayout:checkBuff(unit, name, count, icon, rank, duration)
	local class
	local proceed
	
	_, class = UnitClass(unit)
	
	if self.db.profile.activateStats then
		local _, isSignificant = self:getAssumedSpec(unit)
		if isSignificant then
			-- If Stat Analysis mode is activated and stats are significant for unit then 
			-- the only buffs that are likely to have an effect are the druids cat/bear ones
			if class=="DRUID" and (
				name==BS["Dire Bear Form"] or name==BS["Cat Form"] or name==BS["Bear Form"]
			) then
				proceed = true
			else
				proceed=false
			end
		else
			proceed = true
		end
	else
		proceed = true
	end
	
	if proceed and (
		(
			class=="PRIEST" and name==BS["Shadowform"]
		) or (
			class=="DRUID" and (
				name==BS["Dire Bear Form"] or name==BS["Cat Form"] or name==BS["Moonkin Form"] or name==BS["Bear Form"]
			)
		) or (
			class=="PALADIN" and (
				name==BS["Righteous Fury"] or name==BS["Improved Righteous Fury"]
			)
		)
	) then
		self:setUpdateNeeded("Pertinent buff change found - "..UnitName(unit).." ("..name..")")
	end
end

-- Scans the whole raid (at beginning and end of a fight) and add statistics on
-- the supposed role of each unit (currently supports only PAL / DRU / PRI)
-- Will call setUpdateNeeded if it detects an assumed role change
function GridDynamicLayout:UpdateStatisticalAnalysis()
	-- Is only done if statistical analysis is activated for profile
	if not self.db.profile.activateStats then return end
	
	-- Initializes the list of members whose roles have changed according to statistical analysis (debug purposes)
	local membersRoleUpdated = ""
	-- initializes the updateNeeded boolean (by default no layout update is needed)
	local updateNeeded = false
	
	for i=1,40 do
		local unit = "raid"..i
		if UnitExists(unit) and not UnitIsDeadOrGhost(unit) then
			local name = self:getRealmedName(unit)
			local _, class = UnitClass(unit)
			if class == "DRUID" then
				-- Initializes stats database for unit if necessary
				if not self.stats[name] then
					self.stats[name] = {
						currentSpec = "resto",
						breakdown = {
							feral=0,
							balance=0,
							resto=0,
						},
					}
				end
				if Aura:UnitHasBuff(unit, BS["Dire Bear Form"]) or Aura:UnitHasBuff(unit, BS["Bear Form"]) or Aura:UnitHasBuff(unit, BS["Cat Form"]) then
					self:incrementSpecStat(name, "feral") 				
				elseif Aura:UnitHasBuff(unit, BS["Moonkin Form"]) then
					self:incrementSpecStat(name, "balance") 
				else
					self:incrementSpecStat(name, "resto")
				end
			elseif class == "PRIEST" then
				-- Initializes stats database for unit if necessary
				if not self.stats[name] then
					self.stats[name] = {
						currentSpec = "holy",
						breakdown = {
							shadow=0,
							holy=0,
						},
					}
				end
				if Aura:UnitHasBuff(unit, BS["Shadowform"]) then
					self:incrementSpecStat(name, "shadow") 				
				else
					self:incrementSpecStat(name, "holy")
				end
			elseif class == "PALADIN" then
				-- Initializes stats database for unit if necessary
				if not self.stats[name] then
					self.stats[name] = {
						currentSpec = "holy",
						breakdown = {
							prot=0,
							holy=0,
						},
					}
				end
				if Aura:UnitHasBuff(unit, BS["Righteous Fury"]) or Aura:UnitHasBuff(unit, BS["Improved Righteous Fury"]) then
					self:incrementSpecStat(name, "prot")
				else
					self:incrementSpecStat(name, "holy")
				end
			end
			-- Checks if an update is needed, that is if someone's assumed spec has changed since the last statistical analysis
			if self.stats[name] then
				--[[if name == "Mokhtar" then
			 		self:Debug("  "..self.stats[name].currentSpec.."~="..self:getAssumedSpec(unit))
			 	end]]--
				if self.stats[name].currentSpec ~= self:getAssumedSpec(unit) then
					updateNeeded = true
					membersRoleUpdated = membersRoleUpdated .. name .. ", " 
					self.stats[name].currentSpec = self:getAssumedSpec(unit)
				end
			end
		end
	end
	
	if updateNeeded then self:setUpdateNeeded("Statistical analysis - member role changed ("..membersRoleUpdated..")") end
end

function GridDynamicLayout:incrementSpecStat(name, spec)
	for k,v in pairs(self.stats[name].breakdown) do
		if k==spec then
			self.stats[name].breakdown[k] = self.stats[name].breakdown[k]+2 
		else
			self.stats[name].breakdown[k] = self.stats[name].breakdown[k]-1
		end
		-- spec score has to be between 0 and MAX_SPECSCORE
		if self.stats[name].breakdown[k] > MAX_SPECSCORE then
			self.stats[name].breakdown[k] = MAX_SPECSCORE
		elseif self.stats[name].breakdown[k] < 0 then
			self.stats[name].breakdown[k] = 0
		end
	end
end

function GridDynamicLayout:getAssumedSpec(unit)
	local name = self:getRealmedName(unit)
	if self.stats[name] then
		local spec = ""
		local specScore = -1
		local isSignificant = false
		for k,v in pairs(self.stats[name].breakdown) do
			if v > specScore then
				spec = k
				specScore = v
			end 
		end
		if specScore >= MIN_SIGNIFICANT_SPECSCORE then
			isSignificant = true
		end
		return spec, isSignificant
	else
		return nil, false
	end
end

function GridDynamicLayout:dumpStatsDB()
	self:Debug("----------- DB DUMP -------------")
	for n,s in pairs(self.stats) do
		local output = n .. " ("..s.currentSpec..") : "
		for spec,specScore in pairs(s.breakdown) do
			output = output..spec..">"..specScore..", "
		end
		self:Debug(output)
	end
end

-- Adds dynamic layout to Grid's layouts
GridLayout:AddLayout("Dynamic Layout", dynamicLayout)
